@@ -8,16 +8,28 @@
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
 
+from collections.abc import Iterator
 from enum import Enum
+from random import Random
+from typing import TYPE_CHECKING, Callable, Optional, Union
 
 from sortedcontainers import SortedList
 
-from hypothesis.internal.conjecture.data import ConjectureData, ConjectureResult, Status
-from hypothesis.internal.conjecture.junkdrawer import LazySequenceCopy, swap
-from hypothesis.internal.conjecture.shrinker import sort_key_ir
+from hypothesis.internal.conjecture.choice import choices_key
+from hypothesis.internal.conjecture.data import (
+    ConjectureData,
+    ConjectureResult,
+    Status,
+    _Overrun,
+)
+from hypothesis.internal.conjecture.junkdrawer import LazySequenceCopy
+from hypothesis.internal.conjecture.shrinker import sort_key
 
 NO_SCORE = float("-inf")
 
+if TYPE_CHECKING:
+    from hypothesis.internal.conjecture.engine import ConjectureRunner
+
 
 class DominanceRelation(Enum):
     NO_DOMINANCE = 0
@@ -26,7 +38,7 @@ class DominanceRelation(Enum):
     RIGHT_DOMINATES = 3
 
 
-def dominance(left, right):
+def dominance(left: ConjectureResult, right: ConjectureResult) -> DominanceRelation:
     """Returns the dominance relation between ``left`` and ``right``, according
     to the rules that one ConjectureResult dominates another if and only if it
     is better in every way.
@@ -45,8 +57,8 @@ def dominance(left, right):
     more structured or failing tests it can be useful to track, and future work
     will depend on it more."""
 
-    left_key = sort_key_ir(left.ir_nodes)
-    right_key = sort_key_ir(right.ir_nodes)
+    left_key = sort_key(left.nodes)
+    right_key = sort_key(right.nodes)
     if left_key == right_key:
         return DominanceRelation.EQUAL
 
@@ -124,21 +136,25 @@ class ParetoFront:
     see how much of a problem this is in practice before we try that.
     """
 
-    def __init__(self, random):
+    def __init__(self, random: Random) -> None:
         self.__random = random
-        self.__eviction_listeners = []
+        self.__eviction_listeners: list[Callable[[ConjectureResult], None]] = []
 
-        self.front = SortedList(key=lambda d: sort_key_ir(d.ir_nodes))
-        self.__pending = None
+        self.front: SortedList[ConjectureResult] = SortedList(
+            key=lambda d: sort_key(d.nodes)
+        )
+        self.__pending: Optional[ConjectureResult] = None
 
-    def add(self, data):
+    def add(self, data: Union[ConjectureData, ConjectureResult, _Overrun]) -> bool:
         """Attempts to add ``data`` to the pareto front. Returns True if
         ``data`` is now in the front, including if data is already in the
         collection, and False otherwise"""
         if data.status < Status.VALID:
             return False
 
+        assert not isinstance(data, _Overrun)
         data = data.as_result()
+        assert not isinstance(data, _Overrun)
 
         if not self.front:
             self.front.add(data)
@@ -165,7 +181,7 @@ class ParetoFront:
             # We track which values we are going to remove and remove them all
             # at the end so the shape of the front doesn't change while we're
             # using it.
-            to_remove = []
+            to_remove: list[ConjectureResult] = []
 
             # We now iteratively sample elements from the approximate pareto
             # front to check whether they should be retained. When the set of
@@ -198,7 +214,7 @@ class ParetoFront:
             dominators = [data]
 
             while i >= 0 and len(dominators) < 10:
-                swap(front, i, self.__random.randint(0, i))
+                front.swap(i, self.__random.randint(0, i))
 
                 candidate = front[i]
 
@@ -213,7 +229,7 @@ class ParetoFront:
                             already_replaced = True
                             dominators[j] = candidate
                             j += 1
-                        else:
+                        else:  # pragma: no cover # flaky, by test_database_contains_only_pareto_front
                             dominators[j], dominators[-1] = (
                                 dominators[-1],
                                 dominators[j],
@@ -232,31 +248,36 @@ class ParetoFront:
                 i -= 1
 
             for v in to_remove:
-                self.__remove(v)
+                self._remove(v)
             return data in self.front
         finally:
             self.__pending = None
 
-    def on_evict(self, f):
+    def on_evict(self, f: Callable[[ConjectureResult], None]) -> None:
         """Register a listener function that will be called with data when it
         gets removed from the front because something else dominates it."""
         self.__eviction_listeners.append(f)
 
-    def __contains__(self, data):
-        return isinstance(data, (ConjectureData, ConjectureResult)) and (
-            data.as_result() in self.front
-        )
+    def __contains__(self, data: object) -> bool:
+        if not isinstance(data, (ConjectureData, ConjectureResult)):
+            return False
 
-    def __iter__(self):
+        result = data.as_result()
+        if isinstance(result, _Overrun):
+            return False
+
+        return result in self.front
+
+    def __iter__(self) -> Iterator[ConjectureResult]:
         return iter(self.front)
 
-    def __getitem__(self, i):
+    def __getitem__(self, i: int) -> ConjectureResult:
         return self.front[i]
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.front)
 
-    def __remove(self, data):
+    def _remove(self, data: ConjectureResult) -> None:
         try:
             self.front.remove(data)
         except ValueError:
@@ -276,11 +297,12 @@ class ParetoOptimiser:
     grow more powerful over time.
     """
 
-    def __init__(self, engine):
+    def __init__(self, engine: "ConjectureRunner") -> None:
         self.__engine = engine
-        self.front = self.__engine.pareto_front
+        assert self.__engine.pareto_front is not None
+        self.front: ParetoFront = self.__engine.pareto_front
 
-    def run(self):
+    def run(self) -> None:
         seen = set()
 
         # We iterate backwards through the pareto front, using the shrinker to
@@ -300,7 +322,7 @@ class ParetoOptimiser:
             assert self.front
             i = min(i, len(self.front) - 1)
             target = self.front[i]
-            if target.buffer in seen:
+            if choices_key(target.choices) in seen:
                 i -= 1
                 continue
             assert target is not prev
@@ -320,15 +342,12 @@ class ParetoOptimiser:
                     # must be dominated in the front - either ``destination`` is in
                     # the front, or it was not added to it because it was
                     # dominated by something in it.
-                    try:
-                        self.front.front.remove(source)
-                    except ValueError:
-                        pass
+                    self.front._remove(source)
                     return True
                 return False
 
             shrunk = self.__engine.shrink(target, allow_transition=allow_transition)
-            seen.add(shrunk.buffer)
+            seen.add(choices_key(shrunk.choices))
 
             # Note that the front may have changed shape arbitrarily when
             # we ran the shrinker. If it didn't change shape then this is
